//Jose M. Pea
//CompBio
//15/10/2004

//k-greedy Markov blanket search.

//Improvements:
//If the Markov blankets of all the variables are requested (target=-1), then
//the algorithm can be speeded up a lot by not cleaning cache_pc_superset[][] 
//(cache_pc_superset[][0]=-2) in pc_superset().
//All the memory structures are static and they could be replaced by dynamic ones.
//Some arrays of integers can be replaced by arrays of binary.

//Assumptions: 
//MAX_LINE_LENGTH, MAX_N_STATES, MAX_STATE_LENGTH and MAX_SIZE_CACHE_PC_SUPERSET.

///////////////////////////////////////////////////////////////////////////////
////////////////////////////////////includes///////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <float.h>
#include <math.h>
#include <time.h>

#include "chisq.h"

///////////////////////////////////////////////////////////////////////////////
//////////////////////////////////defines//////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

#define MAX_LINE_LENGTH 100000
#define MAX_N_STATES 10
#define MAX_STATE_LENGTH 100

#define MAX_SIZE_CACHE_PC_SUPERSET 100

#define N_CASES_PER_DF 5

///////////////////////////////////////////////////////////////////////////////
////////////////////////////////global definitions/////////////////////////////
///////////////////////////////////////////////////////////////////////////////

int n_cases, n_vars, true_net, max_max_cond_size;

double k_tradeoff, alpha;

clock_t clock_init;

int *n_states;

char *data_file;

int **data, **net, **cache_pc_superset;

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////functions' prototypes///////////////////////////
///////////////////////////////////////////////////////////////////////////////

void parse_data(void);
int state_index(int var, char *state, char ***states);
void algorithm_mb(int target);
void interiamb(int target);
void pc_set(int target, int cache_pc_superset_allowed, int *pc, int **sep);
void pc_superset(int target, int cache_pc_superset_allowed, int *pc, int **sep);
double min_dep(int var, int target, int compulsory, int *pc, int *sep);
int next_cond_index(int n_pc, int cond_size, int *cond_index);
double compute_dep(int var, int target, int *cond);
int k_greedy(double *dep);
void report(int target, int *pc, int *mb);

///////////////////////////////////////////////////////////////////////////////
////////////////////////////////functions' body////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////

void main(int argc, char* argv[])

{int i, target, algorithm;

if(argc!=9) 
	{printf("\n\n Usage: data_file n_cases n_vars target(-1=all) k_tradeoff true_net(0=no,1=yes) algorithm(0=interiamb,1=algorithm_mb) alpha. \n\n"); 
	 exit(0);
	}

srand((unsigned)time(NULL));

clock_init=clock();

data_file=argv[1];
n_cases=atoi(argv[2]);
n_vars=atoi(argv[3]);
target=atoi(argv[4]);
k_tradeoff=atof(argv[5]);
true_net=atoi(argv[6]);
algorithm=atoi(argv[7]);
alpha=atof(argv[8]);

n_states=new int[n_vars];

data=new int*[n_cases];
for(i=0;i<n_cases;i++)
	data[i]=new int[n_vars];

if(true_net==1)
	{net=new int*[n_vars];
	 for(i=0;i<n_vars;i++)
		net[i]=new int[n_vars];
	}

parse_data();

cache_pc_superset=new int*[n_vars];
for(i=0;i<n_vars;i++)
	{cache_pc_superset[i]=new int[(int)MAX_SIZE_CACHE_PC_SUPERSET];
	 cache_pc_superset[i][0]=-2;
	}

if(algorithm==0)
	{if(target==-1)
		for(i=0;i<n_vars;i++)
			interiamb(i);
	 else
		interiamb(target);
	}
else
	{if(target==-1)
		for(i=0;i<n_vars;i++)
			algorithm_mb(i);
	 else
		algorithm_mb(target);
	}

//printf("\n\n Assumption: MAX_LINE_LENGTH, MAX_N_STATES, MAX_STATE_LENGTH and MAX_SIZE_CACHE_PC_SUPERSET.");
}

/////////////////////////////////////

void parse_data(void)

{int i, j, min_n_states;
 char aux;
 char name2[80];
 char buffer[(int)MAX_LINE_LENGTH];
 char *token;
 const char *separators=" ,\t\n\r";
 FILE *f_in;
 char ***states;

if(true_net==1)
	{sprintf(name2,"%s.net",data_file);
	 f_in=fopen(name2,"r");

	 for(i=0;i<n_vars;i++)
		{for(j=0;j<n_vars;j++)
			{fscanf(f_in,"%c",&aux);
		     net[i][j]=atoi(&aux);
			}

		 fscanf(f_in,"%c",&aux);
		}

	 fclose(f_in);
	}

states=new char**[n_vars];
for(i=0;i<n_vars;i++)
	{states[i]=new char*[(int)MAX_N_STATES];
	 for(j=0;j<(int)MAX_N_STATES;j++)
		states[i][j]=new char[(int)MAX_STATE_LENGTH];
	}

for(i=0;i<n_vars;i++)
	n_states[i]=0;

f_in=fopen(data_file,"r");

for(i=0;i<n_cases;i++)
	{fgets(buffer,(int)MAX_LINE_LENGTH,f_in);
	 if((int)strlen(buffer)==(int)MAX_LINE_LENGTH-1)
		{printf("\n\n Exceeding MAX_LINE_LENGTH. \n\n");
		 exit(0);
		}
  
	 token=strtok(buffer,separators); 
	 for(j=0;j<n_vars;j++)
		{if((int)strlen(token)>(int)MAX_STATE_LENGTH-1)
			{printf("\n\n Exceeding MAX_STATE_LENGTH. \n\n");
			 exit(0);
			}

		 data[i][j]=state_index(j,token,states);
			 
		 token=strtok(0,separators);
		}
	}

fclose(f_in);

min_n_states=0;
for(i=0;i<n_vars;i++)
	if(n_states[i]>1 && (min_n_states==0 || n_states[i]<min_n_states))
		min_n_states=n_states[i];

if(min_n_states==0)
	{printf("\n\n Only one state for all the variables. \n\n");
	 exit(0);
	}

max_max_cond_size=(int)(log((double)n_cases/(double)((int)N_CASES_PER_DF*(min_n_states-1)*(min_n_states-1)))/log((double)min_n_states)); //min_n_states may not be representative.

for(i=0;i<n_vars;i++)
	{for(j=0;j<(int)MAX_N_STATES;j++)
		delete [] states[i][j];
	 delete [] states[i];
	}
delete [] states;
}

/////////////////////////////////////

int state_index(int var, char *state, char ***states)

{int i;

for(i=0;i<n_states[var];i++)
	if(strcmp(states[var][i],state)==0)
		return i;

if(i>(int)MAX_N_STATES-1)
	{printf("\n\n Exceeding MAX_N_STATES. \n\n");
	 exit(0);
	}

strcpy(states[var][i],state);
n_states[var]++;

return i;
}

/////////////////////////////////////

void interiamb(int target)

//interIAMB(target).

{int i, j, k, stop, stop2, n_conds, aux;
 int *pc, *mb, *removals, *cond;
 double *dep;

pc=new int[n_vars];
mb=new int[n_vars];
removals=new int[n_vars];
for(i=0;i<n_vars;i++)
	{pc[i]=0;
	 if(n_states[i]>1 && n_states[target]>1)
		mb[i]=0;
	 else
		mb[i]=-1;
	 removals[i]=0;
	}
mb[target]=-1;

n_conds=0;
cond=new int[max_max_cond_size+1];
for(i=0;i<max_max_cond_size+1;i++)
	cond[i]=-1;

dep=new double[n_vars];

do
	{stop=1; 

	 if(n_conds<=max_max_cond_size)
		{for(i=0;i<n_vars;i++)
			dep[i]=(double)0.0;

		 for(i=0;i<n_vars;i++)
			if(mb[i]==0)
				{dep[i]=compute_dep(i,target,cond);

				 if(dep[i]>=(double)1.0)
					stop=0;
				}
		}

	 if(stop==0)
		{aux=k_greedy(dep);

		 mb[aux]=1;
		 cond[n_conds]=aux;
		 n_conds++;

		 //printf("\n Adding %d with dependence %.2f",aux,dep[aux]);
		}
	 else
		 do
			{stop2=1;

			 for(i=0;i<n_vars;i++)
				dep[i]=(double)0.0;
			
			 for(i=0;i<n_vars;i++)
				if(mb[i]==1 && removals[i]<10)
					{for(j=0;j<n_conds;j++)
						if(cond[j]==i)
							{for(k=j;k<n_conds-1;k++)
								 cond[k]=cond[k+1];
							 cond[n_conds-1]=-1;

							 j=n_conds;
							}

					 dep[i]=compute_dep(i,target,cond);

					 cond[n_conds-1]=i;

					 if(dep[i]<=(double)(-1.0))
						{dep[i]=(double)1.0;
						 stop=0;
						 stop2=0;
						}
					 else
						dep[i]=(double)0.0;
					}

			 if(stop2==0)
				{aux=k_greedy(dep);

				 mb[aux]=0;
				 removals[aux]++; //To avoid endless loop.
				 for(j=0;j<n_conds;j++)
					 if(cond[j]==aux)
						{for(k=j;k<n_conds-1;k++)
							 cond[k]=cond[k+1];
						 cond[n_conds-1]=-1;

						 j=n_conds;
						}
				 n_conds--;

				 //printf("\n Removing %d",aux);
				}
			}
		 while(stop2==0);
	}
while(stop==0);

for(i=0;i<n_vars;i++)
	if(mb[i]==-1)
		mb[i]=0;

report(target,pc,mb);

printf(" nodes with 10 removals: ");
for(i=0;i<n_vars;i++)
	if(removals[i]==10)
		printf("%d ",i);

delete [] pc;
delete [] mb;
delete [] removals;
delete [] cond;
delete [] dep;
}

/////////////////////////////////////

void algorithm_mb(int target)

//AlgorithmMB(target).
//If the Markov blankets of all the variables are requested (target=-1), then the 
//algorithm can be speeded up a lot by not cleaning cache_pc_superset[][] (cache_pc_superset[][0]=-2).
//The memory requirement can be reduced by removing sep2[][], which is not used at all.

{int i, j, k, in_cond, n_conds;
 int *pc, *pc2, *mb, *cond;
 int **sep, **sep2;

pc=new int[n_vars];
pc2=new int[n_vars];
mb=new int[n_vars];

cond=new int[max_max_cond_size];

sep=new int*[n_vars];
sep2=new int*[n_vars];
for(i=0;i<n_vars;i++)
	{sep[i]=new int[max_max_cond_size];
	 sep2[i]=new int[max_max_cond_size];
	}

for(i=0;i<n_vars;i++)
	cache_pc_superset[i][0]=-2;

pc_set(target,0,pc,sep);

for(i=0;i<n_vars;i++)
	mb[i]=pc[i];

for(i=0;i<n_vars;i++)
	if(pc[i]==1)
		{pc_set(i,1,pc2,sep2);

		 for(j=0;j<n_vars;j++)
			if(pc2[j]==1 && pc[j]==0 && j!=target && mb[j]==0)
				{in_cond=0;
				 n_conds=0;
				 for(k=0;k<max_max_cond_size;k++)
					{cond[k]=sep[j][k];

					 if(cond[k]==i)
						in_cond=1;

					 if(cond[k]!=-1)
						n_conds++;
					}

				 if(in_cond==0 && n_conds<max_max_cond_size)
					{cond[n_conds]=i;
				 
					 if(compute_dep(j,target,cond)>=(double)1.0)
						mb[j]=1;
					}
				}
		}

report(target,pc,mb);

delete [] pc;
delete [] pc2;
delete [] mb;
delete [] cond;
for(i=0;i<n_vars;i++)
	{delete [] sep[i];
	 delete [] sep2[i];
	}
delete [] sep;
delete [] sep2;
}

/////////////////////////////////////

void pc_set(int target, int cache_pc_superset_allowed, int *pc, int **sep)

//AlgorithmPC(target). 
//It returns pc[01] and sep[][-1i] (sep[][-1i] is valid iff cache_pc_superset_allowed=0).

{int i, j;
 int *pc2;
 int **sep2;

//printf("\n\n pc_set with target %d",target);

pc2=new int[n_vars];

sep2=new int*[n_vars];
for(i=0;i<n_vars;i++)
	sep2[i]=new int[max_max_cond_size];

pc_superset(target,cache_pc_superset_allowed,pc,sep);

for(i=0;i<n_vars;i++)
	if(pc[i]==1)
		{pc_superset(i,cache_pc_superset_allowed,pc2,sep2);

		 if(pc2[target]==0)
			{pc[i]=0;

			 for(j=0;j<max_max_cond_size;j++)
				sep[i][j]=sep2[target][j];
			}
		}

delete [] pc2;
for(i=0;i<n_vars;i++)
	delete [] sep2[i];
delete [] sep2;
}

/////////////////////////////////////

void pc_superset(int target, int cache_pc_superset_allowed, int *pc, int **sep)

//AlgorithmPCD(target).
//It returns pc[01] and sep[][-1i] (sep[][-1i] is valid iff cache_pc_superset_allowed=0).

{int i, j, last_added, stop;
 int *sep2;
 double *dep, *cache_dep;

//printf("\n\n pc_superset with target %d",target);

for(i=0;i<n_vars;i++)
	for(j=0;j<max_max_cond_size;j++)
		sep[i][j]=-1;

if(cache_pc_superset_allowed==1 && cache_pc_superset[target][0]!=-2)
	{for(i=0;i<n_vars;i++)
		pc[i]=0;

	 for(i=0;i<(int)MAX_SIZE_CACHE_PC_SUPERSET;i++)
		if(cache_pc_superset[target][i]!=-1)
			pc[cache_pc_superset[target][i]]=1;

	 return;
	}

sep2=new int[max_max_cond_size];

dep=new double[n_vars];

cache_dep=new double[n_vars];
for(i=0;i<n_vars;i++)
	cache_dep[i]=(double)0.0;

for(i=0;i<n_vars;i++)
	if(n_states[i]>1 && n_states[target]>1)
		pc[i]=0;
	else
		pc[i]=-1;
pc[target]=-1;

last_added=-1;

do
	{stop=1; 

	 for(i=0;i<n_vars;i++)
		dep[i]=(double)0.0;

	 for(i=0;i<n_vars;i++)
		if(pc[i]==0)
			{dep[i]=min_dep(i,target,last_added,pc,sep2);

			 if(dep[i]==(double)0.0)
				dep[i]=cache_dep[i];
			 else
				{if(cache_dep[i]==(double)0.0 || dep[i]<cache_dep[i])
					{cache_dep[i]=dep[i];

					 if(dep[i]<=(double)(-1.0))
						for(j=0;j<max_max_cond_size;j++)
							sep[i][j]=sep2[j];
					}
				 else
					dep[i]=cache_dep[i];
				}
			}

	 for(i=0;i<n_vars;i++)
		if(dep[i]<=(double)(-1.0))
			pc[i]=-1;
		else
			if(dep[i]>=(double)1.0)
				stop=0;

	 if(stop==0)
		{last_added=k_greedy(dep);

		 pc[last_added]=1;

		 //printf("\n Adding %d with dependence %.2f",last_added,dep[last_added]);
		}
	 else
		{for(i=0;i<n_vars;i++)
			dep[i]=(double)0.0;
	
		 for(i=0;i<n_vars;i++)
			if(pc[i]==1 && i!=last_added)
				{pc[i]=0; 

				 dep[i]=min_dep(i,target,last_added,pc,sep2);

				 pc[i]=1;

				 if(dep[i]<=(double)(-1.0))
					for(j=0;j<max_max_cond_size;j++)
						sep[i][j]=sep2[j];
				}

		 for(i=0;i<n_vars;i++)
			if(pc[i]==1 && i!=last_added && dep[i]<=(double)(-1.0))
				{pc[i]=-1;

				 stop=0;

				 //printf("\n Removing %d",i);
				}
		}
	}
while(stop==0);

for(i=0;i<n_vars;i++)
	if(pc[i]==-1)
		pc[i]=0;

for(i=0;i<(int)MAX_SIZE_CACHE_PC_SUPERSET;i++)
	cache_pc_superset[target][i]=-1;

j=0;
for(i=0;i<n_vars;i++)
	if(pc[i]==1)
		{if(j>(int)MAX_SIZE_CACHE_PC_SUPERSET)
			{printf("\n\n Exceeding MAX_SIZE_CACHE_PC_SUPERSET. \n\n");
			 exit(0);
			}
	
		 cache_pc_superset[target][j]=i;

		 j++;
		}

delete [] sep2;
delete [] dep;
delete [] cache_dep;
}

/////////////////////////////////////

double min_dep(int var, int target, int compulsory, int *pc, int *sep)

//It computes the minimum of the depence measure between var and target conditioned on 
//subsets of pc[x1] st compulsory is always included (-1=no compulsory). Note that only
//those subsets of pc[x1] that have a size no larger than max_cond_size are considered.
//It returns the minimum dependence [-inf,-1] for H0 (the smaller the more independent), 
//[1,+inf] for H1 (the higher the more dependent), or 0 for not enough data (see N_CASES_PER_DF),
//and sep[-1i].

{int i, j, n_pc, stop, cond_size, max_cond_size;
 double dep, aux;
 int *cond, *code, *cond_index;

cond=new int[max_max_cond_size];
for(i=0;i<max_max_cond_size;i++)
	{cond[i]=-1;
	 sep[i]=-1;
	}

n_pc=0;
for(i=0;i<n_vars;i++)
	if(pc[i]==1 && i!=compulsory)
		n_pc++;

max_cond_size=max_max_cond_size;
if(compulsory!=-1)
	max_cond_size--;
if(max_cond_size>n_pc)
	max_cond_size=n_pc;

code=new int[n_pc];
j=0;
for(i=0;i<n_vars;i++)
	if(pc[i]==1 && i!=compulsory)
		{code[j]=i;

		 j++;
		}

dep=(double)0.0;
for(cond_size=0;cond_size<=max_cond_size;cond_size++)
	{cond_index=new int[cond_size];
	 for(i=0;i<cond_size;i++)
		cond_index[i]=i;

	 do
		{stop=0;
		
		 for(i=0;i<cond_size;i++)
			cond[i]=code[cond_index[i]];
		 if(compulsory!=-1)
			cond[cond_size]=compulsory;
	
		 aux=compute_dep(var,target,cond);

		 if(aux!=(double)0.0 && (dep==(double)0.0 || aux<dep))
			{dep=aux;
			 
			 if(dep<=(double)(-1.0))
				{for(i=0;i<max_max_cond_size;i++)
					sep[i]=cond[i];

				 stop=1;
				 cond_size=max_cond_size+1;
				}
			}
	 
		 if(stop==0)
			stop=next_cond_index(n_pc,cond_size,cond_index);
		}
	 while(stop==0);

	 delete [] cond_index;
	}

delete [] code;
delete [] cond;

return dep;
}

/////////////////////////////////////

int next_cond_index(int n_pc, int cond_size, int *cond_index)

//Based on the function subSetsOfSize() from the Elvira project.

{int i, j, stop;

stop=1;
for(i=cond_size-1;i>=0;i--)
	{if(cond_index[i]<n_pc+i-cond_size)
		{cond_index[i]++;
		 
		 if(i<cond_size-1)
			for(j=i+1;j<cond_size;j++)
				cond_index[j]=cond_index[j-1]+1;

		 stop=0;
		 i=-1;
		}
	}

return stop;
}

/////////////////////////////////////

double compute_dep(int var, int target, int *cond)

//It computes the depence measure between var and target conditioned on cond[-1i], where
//cond[-1i] is an array of size max_max_cond_size.
//It returns the dependence [-inf,-1] for H0 (the smaller the more independent), [+1,+inf]
//for H1 (the higher the more dependent), or 0 for not enough data (see N_CASES_PER_DF).

//It can be speeded up by allocating memory only once in main() at the cost of memory space, 
//and by caching the critical value of the test as a function of the degrees of freedom.
//Only ss[][][] is strictly necessary. The rest of the arrays are used to gain speed at the 
//cost of memory space.

{int i, j, k, n_cond_states, cond_state, df, df_target, df_var;
 double statistic, dep, aux;
 int *ss_cond;
 int **ss_target, **ss_var;
 int ***ss;

n_cond_states=1;
for(i=0;i<max_max_cond_size;i++) 
	if(cond[i]!=-1)
		n_cond_states=n_cond_states*n_states[cond[i]];

if(n_cond_states*(n_states[target]-1)*(n_states[var]-1)*(int)N_CASES_PER_DF>n_cases)
	return (double)0.0;

ss_cond=new int[n_cond_states];
for(i=0;i<n_cond_states;i++)
	ss_cond[i]=0;

ss_target=new int*[n_cond_states];
for(i=0;i<n_cond_states;i++)
	{ss_target[i]=new int[n_states[target]];
	 for(j=0;j<n_states[target];j++)
		ss_target[i][j]=0;
	}

ss_var=new int*[n_cond_states];
for(i=0;i<n_cond_states;i++)
	{ss_var[i]=new int[n_states[var]];
	 for(j=0;j<n_states[var];j++)
		ss_var[i][j]=0;
	}

ss=new int**[n_cond_states];
for(i=0;i<n_cond_states;i++)
	{ss[i]=new int*[n_states[target]];
	 for(j=0;j<n_states[target];j++)
		{ss[i][j]=new int[n_states[var]];
		 for(k=0;k<n_states[var];k++)
			 ss[i][j][k]=0;
		}
	}

for(i=0;i<n_cases;i++)
	{cond_state=0;
	 for(j=0;j<max_max_cond_size;j++)
		if(cond[j]!=-1)
			cond_state=cond_state*n_states[cond[j]]+data[i][cond[j]];

	 ss_cond[cond_state]++;
	 ss_target[cond_state][data[i][target]]++;
	 ss_var[cond_state][data[i][var]]++;
	 ss[cond_state][data[i][target]][data[i][var]]++;
	}

//X^2 statistic.
/*
statistic=(double)0.0;
for(i=0;i<n_cond_states;i++)
	if(ss_cond[i]>0)
		for(j=0;j<n_states[target];j++)
			if(ss_target[i][j]>0)
				for(k=0;k<n_states[var];k++)
					if(ss_var[i][k]>0)
						{aux=(double)(ss_target[i][j]*ss_var[i][k])/(double)ss_cond[i];
						 statistic=statistic+pow((double)ss[i][j][k]-aux,(double)2.0)/aux;
						}
*/

//G^2 statistic based on mutual information.
/*
statistic=(double)0.0;
for(i=0;i<n_cond_states;i++)
	if(ss_cond[i]>0)
		for(j=0;j<n_states[target];j++)
			if(ss_target[i][j]>0)
				for(k=0;k<n_states[var];k++)
					if(ss[i][j][k]>0)
						statistic=statistic+(double)ss[i][j][k]*(log10((double)ss[i][j][k])-
						log10((double)ss_target[i][j])-log10((double)ss_var[i][k])+
						log10((double)ss_cond[i]))/log10((double)2.0);
statistic=statistic*(double)2.0;
*/

//G^2 statistic based on observed and expected frequencies.

statistic=(double)0.0;
for(i=0;i<n_cond_states;i++)
	if(ss_cond[i]>0)
		for(j=0;j<n_states[target];j++)
			if(ss_target[i][j]>0)
				for(k=0;k<n_states[var];k++)
					if(ss[i][j][k]>0)
						{aux=(double)(ss_target[i][j]*ss_var[i][k])/(double)ss_cond[i];
						 statistic=statistic+(double)ss[i][j][k]*(log((double)ss[i][j][k])-log(aux));
						}		
statistic=statistic*(double)2.0;


//Reduced df due to zero entries. This does not make sense with the X^2 statistic.
/*
df=n_cond_states*(n_states[target]-1)*(n_states[var]-1);
for(i=0;i<n_cond_states;i++)
	for(j=0;j<n_states[target];j++)
		for(k=0;k<n_states[var];k++)
			if(ss[i][j][k]==0)
				df--;
*/

//Reduced df due to zero marginals.

df=0;
for(i=0;i<n_cond_states;i++)
	if(ss_cond[i]>0)
		{df_target=0;
		 for(j=0;j<n_states[target];j++)
			if(ss_target[i][j]>0)
				df_target++;

		 df_var=0;
		 for(k=0;k<n_states[var];k++)
			 if(ss_var[i][k]>0)
				df_var++;

		 df=df+(df_target-1)*(df_var-1);
		}


delete [] ss_cond;
for(i=0;i<n_cond_states;i++)
	{delete [] ss_target[i];
	 delete [] ss_var[i];
	 for(j=0;j<n_states[target];j++)
		delete [] ss[i][j];
	 delete [] ss[i];
	}
delete [] ss_target;
delete [] ss_var;
delete [] ss;

if(statistic<(double)0.0) //statistic sometimes takes values -0.0. Due to loss of precision ?
	statistic=(double)0.0;

if(df<=0) //Naive ?
	df=1;

dep=gammq((double)0.5*(double)df,(double)0.5*statistic);

if(dep<(double)0.0) //Just in case.
	dep=(double)0.0;
else
	if(dep>(double)1.0)
		dep=(double)1.0;

if(dep<=alpha)
	{dep=(double)2.0-dep;
	
	 if(dep==(double)2.0)
		dep=(double)2.0+statistic/(double)df;
	
	 return dep;
	}
else
	{dep=(double)(-1.0)-dep;
	
	 if(dep==(double)(-2.0))
		dep=(double)(-2.0)-statistic/(double)df;
	
	 return dep;
	}
}

/////////////////////////////////////

int k_greedy(double *dep)

//k-greedy (maximization) over the nodes with dep[] >= 1.0.
//It returns the index of the maximum.

{int i, j, k, n_cands, n_cands_left, max;
 int *cand;

n_cands=0;
for(i=0;i<n_vars;i++)
	if(dep[i]>=(double)1.0)
		n_cands++;

cand=new int[n_cands];

j=0;
for(i=0;i<n_vars;i++)
	if(dep[i]>=(double)1.0)
		{cand[j]=i;

		 j++;
		}

n_cands_left=n_cands;

n_cands=(int)((double)n_cands*k_tradeoff);
if(n_cands<1)
	n_cands=1;

for(i=0;i<n_cands;i++)
	{j=rand()%n_cands_left;

	 if(i==0 || dep[cand[j]]>dep[max])
		max=cand[j];

	 for(k=j;k<n_cands_left-1;k++)
		cand[k]=cand[k+1];
	 
	 n_cands_left--;
	}

delete [] cand;

return max;
}

/////////////////////////////////////

void report(int target, int *pc, int *mb)

{int i, j, in_mb, tp, fp, fn;
 		
printf("\n");

if(true_net==1)
	{tp=0;
	 fp=0;
	 fn=0;

	 for(i=0;i<n_vars;i++)
		if(i!=target)
			{in_mb=0;
			
			 if(net[i][target]==1 || net[target][i]==1)
				 in_mb=1;
			 else
				 for(j=0;j<n_vars;j++)
					 if(net[target][j]==1 && net[i][j]==1)
						 in_mb=1;

			 if(mb[i]==1 && in_mb==1)
				tp++;
			 else
				{if(mb[i]==1 && in_mb==0)
					fp++;
				 else
					if(mb[i]==0 && in_mb==1)
						fn++;
				}
			}

	 if(tp==0 && fp==0)
		printf("%.2f",(double)(-1.0));
	 else
		printf("%.2f",(double)tp/(double)(tp+fp));

	 if(tp==0 && fn==0)
		printf(" %.2f",(double)(-1.0));
	 else
		printf(" %.2f",(double)tp/(double)(tp+fn));

	 if((tp==0 && fp==0) || (tp==0 && fn==0))
		printf(" %.2f",(double)(-1.0));
	 else
		printf(" %.2f",sqrt(pow((double)1.0-(double)tp/(double)(tp+fp),(double)2.0)+pow((double)1.0-(double)tp/(double)(tp+fn),(double)2.0)));
	}

printf(" %d",(int)(clock()-clock_init)/(int)CLOCKS_PER_SEC);

printf(" pc:");
for(i=0;i<n_vars;i++)
	if(pc[i]==1)
		printf(" %d,",i);
printf(" mb:");
for(i=0;i<n_vars;i++)
	if(mb[i]==1)
		printf(" %d,",i);
}
